from .utils import _NEEDS_EARLY_REACTOR

if _NEEDS_EARLY_REACTOR:
    from scrapy.utils.reactor import install_reactor

    install_reactor("twisted.internet.asyncioreactor.AsyncioSelectorReactor")

from ._downloader_middleware import ScrapyZyteAPIDownloaderMiddleware  # NOQA
from ._request_fingerprinter import ScrapyZyteAPIRequestFingerprinter  # NOQA
from .handler import ScrapyZyteAPIDownloadHandler  # NOQA



from http.cookiejar import Cookie
from typing import Any, Dict, List, Optional
from urllib.parse import urlparse

from scrapy.http import Request
from scrapy.http.cookies import CookieJar


def _get_cookie_jar(request: Request, cookie_jars: Dict[Any, CookieJar]) -> CookieJar:
    jar_id = request.meta.get("cookiejar")
    return cookie_jars[jar_id]


def _get_cookie_domain(cookie, url):
    domain = cookie.get("domain")
    if domain:
        return domain
    domain = urlparse(url).hostname
    if domain:
        return domain
    raise ValueError(
        f"Got a cookie without a domain from URL {url} which has no domain either."
    )


def _process_cookies(
    api_response: Dict[str, Any],
    request: Request,
    cookie_jars: Optional[Dict[Any, CookieJar]],
):
    if not cookie_jars:
        return
    response_cookies = api_response.get("experimental", {}).get("responseCookies")
    if not response_cookies:
        return
    cookie_jar = _get_cookie_jar(request, cookie_jars)
    for response_cookie in response_cookies:
        rest = {}
        http_only = response_cookie.get("httpOnly", None)
        if http_only is not None:
            rest["httpOnly"] = http_only
        same_site = response_cookie.get("sameSite", None)
        if same_site is not None:
            rest["sameSite"] = same_site
        cookie = Cookie(
            version=1,
            name=response_cookie["name"],
            value=response_cookie["value"],
            port=None,
            port_specified=False,
            domain=_get_cookie_domain(response_cookie, api_response["url"]),
            domain_specified="domain" in response_cookie,
            domain_initial_dot=response_cookie.get("domain", "").startswith("."),
            path=response_cookie.get("path", "/"),
            path_specified="path" in response_cookie,
            secure=response_cookie.get("secure", False),
            expires=response_cookie.get("expires", None),
            discard=False,
            comment=None,
            comment_url=None,
            rest=rest,
        )
        cookie_jar.set_cookie(cookie)


def _get_all_cookies(
    request: Request, cookie_jars: Dict[Any, CookieJar]
) -> List[Cookie]:
    cookie_jar = _get_cookie_jar(request, cookie_jars)
    return list(cookie_jar.jar)



from base64 import b64decode
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union

from scrapy import Request
from scrapy.http import HtmlResponse, Response, TextResponse
from scrapy.http.cookies import CookieJar
from scrapy.responsetypes import responsetypes

from scrapy_zyte_api._cookies import _process_cookies
from scrapy_zyte_api.utils import (
    _RESPONSE_HAS_ATTRIBUTES,
    _RESPONSE_HAS_IP_ADDRESS,
    _RESPONSE_HAS_PROTOCOL,
)

_DEFAULT_ENCODING = "utf-8"


class ZyteAPIMixin:
    REMOVE_HEADERS = {
        # Zyte API already decompresses the HTTP Response Body. Scrapy's
        # HttpCompressionMiddleware will error out when it attempts to
        # decompress an already decompressed body based on this header.
        "content-encoding"
    }

    def __init__(self, *args, raw_api_response: Optional[Dict] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._raw_api_response = raw_api_response
        if not _RESPONSE_HAS_ATTRIBUTES:
            self.attributes: Tuple[str, ...] = (
                "url",
                "status",
                "headers",
                "body",
                "request",
                "flags",
                "certificate",
            )
            if _RESPONSE_HAS_IP_ADDRESS:
                self.attributes += ("ip_address",)
            if _RESPONSE_HAS_PROTOCOL:
                self.attributes += ("protocol",)
        self.attributes += ("raw_api_response",)

    def replace(self, *args, **kwargs):
        if kwargs.get("raw_api_response"):
            raise ValueError("Replacing the value of 'raw_api_response' isn't allowed.")
        for attribute in self.attributes:
            kwargs.setdefault(attribute, getattr(self, attribute))
        cls = kwargs.pop("cls", self.__class__)
        return cls(*args, **kwargs)

    @property
    def raw_api_response(self) -> Optional[Dict]:
        """Contains the raw API response from Zyte API.

        To see the full list of parameters and their description, kindly refer to the
        `Zyte API Specification <https://docs.zyte.com/zyte-api/openapi.html#zyte-openapi-spec>`_.
        """
        return self._raw_api_response

    @staticmethod
    def _response_cookie_to_header_value(cookie):
        result = f"{cookie['name']}={cookie['value']}"
        domain = cookie.get("domain")
        if domain:
            result += f"; Domain={cookie['domain']}"
        path = cookie.get("path")
        if path is not None:
            result += f"; Path={path}"
        expires = cookie.get("expires")
        if expires is not None:
            expires_date = datetime.utcfromtimestamp(expires)
            expires_date_string = expires_date.strftime("%a, %d %b %Y %H:%M:%S GMT")
            result += f"; Expires={expires_date_string}"
        if cookie.get("httpOnly"):
            result += "; HttpOnly"
        if cookie.get("secure"):
            result += "; Secure"
        same_site = cookie.get("sameSite")
        if same_site is not None:
            result += f"; SameSite={same_site}"
        return result

    @classmethod
    def _prepare_headers(cls, api_response: Dict[str, Any]):
        result: Dict[str, List[str]] = {}
        input_headers: Optional[List[Dict[str, str]]] = api_response.get(
            "httpResponseHeaders"
        )
        if input_headers:
            result = {
                h["name"]: [h["value"]]
                for h in input_headers
                if h["name"].lower() not in cls.REMOVE_HEADERS
            }
        input_cookies: Optional[List[Dict[str, str]]] = api_response.get(
            "experimental", {}
        ).get("responseCookies")
        if input_cookies:
            result["Set-Cookie"] = []
            for cookie in input_cookies:
                result["Set-Cookie"].append(
                    cls._response_cookie_to_header_value(cookie)
                )
        return result


class ZyteAPITextResponse(ZyteAPIMixin, HtmlResponse):
    @classmethod
    def from_api_response(cls, api_response: Dict, *, request: Request = None):
        """Alternative constructor to instantiate the response from the raw
        Zyte API response.
        """
        body = None
        encoding = None

        if api_response.get("browserHtml"):
            encoding = _DEFAULT_ENCODING  # Zyte API has "utf-8" by default
            body = api_response["browserHtml"].encode(encoding)
        elif api_response.get("httpResponseBody"):
            body = b64decode(api_response["httpResponseBody"])

        return cls(
            url=api_response["url"],
            status=api_response.get("statusCode") or 200,
            body=body,
            encoding=encoding,
            request=request,
            flags=["zyte-api"],
            headers=cls._prepare_headers(api_response),
            raw_api_response=api_response,
        )

    def replace(self, *args, **kwargs):
        kwargs.setdefault("encoding", self.encoding)
        return ZyteAPIMixin.replace(self, *args, **kwargs)


class ZyteAPIResponse(ZyteAPIMixin, Response):
    @classmethod
    def from_api_response(cls, api_response: Dict, *, request: Request = None):
        """Alternative constructor to instantiate the response from the raw
        Zyte API response.
        """
        return cls(
            url=api_response["url"],
            status=api_response.get("statusCode") or 200,
            body=b64decode(api_response.get("httpResponseBody") or ""),
            request=request,
            flags=["zyte-api"],
            headers=cls._prepare_headers(api_response),
            raw_api_response=api_response,
        )


_IMMUTABLE_JSON = Union[None, str, int, float, bool]
_JSON = Union[
    None, str, int, float, bool, List["_JSON"], Dict[_IMMUTABLE_JSON, "_JSON"]
]
_API_RESPONSE = Dict[str, _JSON]


def _process_response(
    api_response: _API_RESPONSE,
    request: Request,
    cookie_jars: Optional[Dict[Any, CookieJar]],
) -> Optional[Union[ZyteAPITextResponse, ZyteAPIResponse]]:
    """Given a Zyte API Response and the ``scrapy.Request`` that asked for it,
    this returns either a ``ZyteAPITextResponse`` or ``ZyteAPIResponse`` depending
    on which if it can properly decode the HTTP Body or have access to browserHtml.
    """

    # NOTES: Currently, Zyte API does NOT only allow both 'browserHtml' and
    # 'httpResponseBody' to be present at the same time. The support for both
    # will be addressed in the future. Reference:
    # - https://github.com/scrapy-plugins/scrapy-zyte-api/pull/10#issuecomment-1131406460
    # For now, at least one of them should be present.

    _process_cookies(api_response, request, cookie_jars)

    if api_response.get("browserHtml"):
        # Using TextResponse because browserHtml always returns a browser-rendered page
        # even when requesting files (like images)
        return ZyteAPITextResponse.from_api_response(api_response, request=request)

    if api_response.get("httpResponseHeaders") and api_response.get("httpResponseBody"):
        response_cls = responsetypes.from_args(
            headers=api_response["httpResponseHeaders"],
            url=api_response["url"],
            # FIXME: update this when python-zyte-api supports base64 decoding
            body=b64decode(api_response["httpResponseBody"]),  # type: ignore
        )
        if issubclass(response_cls, TextResponse):
            return ZyteAPITextResponse.from_api_response(api_response, request=request)

    return ZyteAPIResponse.from_api_response(api_response, request=request)



import json
import logging
from copy import deepcopy
from typing import Generator, Optional, Union

from scrapy import Spider, signals
from scrapy.core.downloader.handlers.http import HTTPDownloadHandler
from scrapy.crawler import Crawler
from scrapy.exceptions import NotConfigured
from scrapy.http import Request
from scrapy.settings import Settings
from scrapy.utils.defer import deferred_from_coro
from scrapy.utils.misc import load_object
from scrapy.utils.reactor import verify_installed_reactor
from twisted.internet.defer import Deferred, inlineCallbacks
from zyte_api.aio.client import AsyncClient, create_session
from zyte_api.aio.errors import RequestError
from zyte_api.apikey import NoApiKey
from zyte_api.constants import API_URL

from ._params import _ParamParser
from .responses import ZyteAPIResponse, ZyteAPITextResponse, _process_response

logger = logging.getLogger(__name__)


def _truncate_str(obj, index, text, limit):
    if len(text) <= limit:
        return
    obj[index] = text[: limit - 1] + "…"


def _truncate(obj, limit):
    if isinstance(obj, dict):
        for key, value in obj.items():
            if isinstance(value, str):
                _truncate_str(obj, key, value, limit)
            elif isinstance(value, (list, dict)):
                _truncate(value, limit)
    elif isinstance(obj, list):
        for index, value in enumerate(obj):
            if isinstance(value, str):
                _truncate_str(obj, index, value, limit)
            elif isinstance(value, (list, dict)):
                _truncate(value, limit)


def _load_retry_policy(settings):
    policy = settings.get("ZYTE_API_RETRY_POLICY")
    if policy:
        policy = load_object(policy)
    return policy


class ScrapyZyteAPIDownloadHandler(HTTPDownloadHandler):
    def __init__(
        self, settings: Settings, crawler: Crawler, client: AsyncClient = None
    ):
        super().__init__(settings=settings, crawler=crawler)
        if not settings.getbool("ZYTE_API_ENABLED", True):
            raise NotConfigured(
                "Zyte API is disabled. Set ZYTE_API_ENABLED to True to enable it."
            )
        if not hasattr(crawler, "zyte_api_client"):
            if not client:
                client = self._build_client(settings)
            # We keep the client in the crawler object to prevent multiple,
            # duplicate clients with the same settings to be used.
            # https://github.com/scrapy-plugins/scrapy-zyte-api/issues/58
            crawler.zyte_api_client = client
        self._client: AsyncClient = crawler.zyte_api_client
        logger.info("Using a Zyte API key starting with %r", self._client.api_key[:7])
        verify_installed_reactor(
            "twisted.internet.asyncioreactor.AsyncioSelectorReactor"
        )
        self._cookies_enabled = settings.getbool("COOKIES_ENABLED")
        self._cookie_jars = None
        self._cookie_mw_cls = load_object(
            settings.get(
                "ZYTE_API_COOKIE_MIDDLEWARE",
                "scrapy.downloadermiddlewares.cookies.CookiesMiddleware",
            )
        )
        self._param_parser = _ParamParser(crawler)
        self._retry_policy = _load_retry_policy(settings)
        self._stats = crawler.stats
        self._session = create_session(
            connection_pool_size=self._client.n_conn,
            trust_env=settings.getbool("ZYTE_API_USE_ENV_PROXY"),
        )
        self._must_log_request = settings.getbool("ZYTE_API_LOG_REQUESTS", False)
        self._truncate_limit = settings.getint("ZYTE_API_LOG_REQUESTS_TRUNCATE", 64)
        if self._truncate_limit < 0:
            raise ValueError(
                f"The value of the ZYTE_API_LOG_REQUESTS_TRUNCATE setting "
                f"({self._truncate_limit}) is invalid. It must be 0 or a "
                f"positive integer."
            )
        crawler.signals.connect(self.engine_started, signal=signals.engine_started)
        if not hasattr(self, "_crawler"):  # Scrapy 2.1 and earlier
            self._crawler = crawler

    def engine_started(self):
        if not self._cookies_enabled:
            return
        for middleware in self._crawler.engine.downloader.middleware.middlewares:
            if isinstance(middleware, self._cookie_mw_cls):
                self._cookie_jars = middleware.jars
                self._param_parser._cookie_jars = self._cookie_jars
                return
        middleware_path = (
            f"{self._cookie_mw_cls.__module__}.{self._cookie_mw_cls.__qualname__}"
        )
        raise RuntimeError(
            f"Could not find a configured downloader middleware that is an "
            f"instance of {middleware_path} (see ZYTE_API_COOKIE_MIDDLEWARE)."
        )

    @staticmethod
    def _build_client(settings):
        try:
            return AsyncClient(
                # To allow users to have a key defined in Scrapy settings and
                # in a environment variable, and be able to cause the
                # environment variable to be used instead of the setting by
                # overriding the setting on the command-line to be an empty
                # string, we do not support setting empty string keys through
                # settings.
                api_key=settings.get("ZYTE_API_KEY") or None,
                api_url=settings.get("ZYTE_API_URL") or API_URL,
                n_conn=settings.getint("CONCURRENT_REQUESTS"),
            )
        except NoApiKey:
            logger.warning(
                "'ZYTE_API_KEY' must be set in the spider settings or env var "
                "in order for ScrapyZyteAPIDownloadHandler to work."
            )
            raise NotConfigured(
                "Your Zyte API key is not set. Set ZYTE_API_KEY to your API key."
            )

    def download_request(self, request: Request, spider: Spider) -> Deferred:
        api_params = self._param_parser.parse(request)
        if api_params is not None:
            return deferred_from_coro(
                self._download_request(api_params, request, spider)
            )
        return super().download_request(request, spider)

    def _update_stats(self):
        prefix = "scrapy-zyte-api"
        for stat in (
            "429",
            "attempts",
            "errors",
            "fatal_errors",
            "processed",
            "success",
        ):
            self._stats.set_value(
                f"{prefix}/{stat}",
                getattr(self._client.agg_stats, f"n_{stat}"),
            )
        for stat in (
            "error_ratio",
            "success_ratio",
            "throttle_ratio",
        ):
            self._stats.set_value(
                f"{prefix}/{stat}",
                getattr(self._client.agg_stats, stat)(),
            )
        for source, target in (
            ("connect", "connection"),
            ("total", "response"),
        ):
            self._stats.set_value(
                f"{prefix}/mean_{target}_seconds",
                getattr(self._client.agg_stats, f"time_{source}_stats").mean(),
            )

        for error_type, count in self._client.agg_stats.api_error_types.items():
            error_type = error_type or "/<empty>"
            if not error_type.startswith("/"):
                error_type = f"/{error_type}"
            self._stats.set_value(f"{prefix}/error_types{error_type}", count)

        for counter in (
            "exception_types",
            "status_codes",
        ):
            for key, value in getattr(self._client.agg_stats, counter).items():
                self._stats.set_value(f"{prefix}/{counter}/{key}", value)

    async def _download_request(
        self, api_params: dict, request: Request, spider: Spider
    ) -> Optional[Union[ZyteAPITextResponse, ZyteAPIResponse]]:
        # Define url by default
        retrying = request.meta.get("zyte_api_retry_policy")
        if retrying:
            retrying = load_object(retrying)
        else:
            retrying = self._retry_policy
        self._log_request(api_params)
        try:
            api_response = await self._client.request_raw(
                api_params,
                session=self._session,
                retrying=retrying,
            )
        except RequestError as er:
            error_detail = (er.parsed.data or {}).get("detail", er.message)
            logger.error(
                f"Got Zyte API error (status={er.status}, type={er.parsed.type!r}) "
                f"while processing URL ({request.url}): {error_detail}"
            )
            raise
        except Exception as er:
            logger.error(
                f"Got an error when processing Zyte API request ({request.url}): {er}"
            )
            raise
        finally:
            self._update_stats()

        return _process_response(api_response, request, self._cookie_jars)

    def _log_request(self, params):
        if not self._must_log_request:
            return
        params = self._truncate_params(params)
        logger.debug(f"Sending Zyte API extract request: {json.dumps(params)}")

    def _truncate_params(self, params):
        if self._truncate_limit == 0:
            return params
        params = deepcopy(params)
        _truncate(params, self._truncate_limit)
        return params

    @inlineCallbacks
    def close(self) -> Generator:
        yield super().close()
        yield deferred_from_coro(self._close())

    async def _close(self) -> None:  # NOQA
        await self._session.close()



import logging

from scrapy.exceptions import IgnoreRequest

from ._params import _ParamParser

logger = logging.getLogger(__name__)


class ScrapyZyteAPIDownloaderMiddleware:
    _slot_prefix = "zyte-api@"

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler)

    def __init__(self, crawler) -> None:
        self._param_parser = _ParamParser(crawler, cookies_enabled=False)
        self._crawler = crawler

        self._max_requests = crawler.settings.getint("ZYTE_API_MAX_REQUESTS")
        if self._max_requests:
            logger.info(
                f"Maximum Zyte API requests for this crawl is set at "
                f"{self._max_requests}. The spider will close when it's "
                f"reached."
            )

    def process_request(self, request, spider):
        if self._param_parser.parse(request) is None:
            return

        downloader = self._crawler.engine.downloader
        slot_id = downloader._get_slot_key(request, spider)
        if not isinstance(slot_id, str) or not slot_id.startswith(self._slot_prefix):
            slot_id = f"{self._slot_prefix}{slot_id}"
            request.meta["download_slot"] = slot_id
        _, slot = downloader._get_slot(request, spider)
        slot.delay = 0

        if self._max_requests_reached(downloader):
            self._crawler.engine.close_spider(spider, "closespider_max_zapi_requests")
            raise IgnoreRequest(
                f"The request {request} is skipped as {self._max_requests} max "
                f"Zyte API requests have been reached."
            )

    def _max_requests_reached(self, downloader) -> bool:
        if not self._max_requests:
            return False

        zapi_req_count = self._crawler.stats.get_value("scrapy-zyte-api/processed", 0)
        download_req_count = sum(
            [
                len(slot.transferring)
                for slot_id, slot in downloader.slots.items()
                if slot_id.startswith(self._slot_prefix)
            ]
        )
        total_requests = zapi_req_count + download_req_count
        return total_requests >= self._max_requests



from typing import TYPE_CHECKING

try:
    from scrapy.utils.request import RequestFingerprinter  # NOQA
except ImportError:
    if not TYPE_CHECKING:
        ScrapyZyteAPIRequestFingerprinter = None
else:
    import hashlib
    import json
    from weakref import WeakKeyDictionary

    from scrapy import Request
    from scrapy.settings.default_settings import REQUEST_FINGERPRINTER_CLASS
    from scrapy.utils.misc import create_instance, load_object
    from w3lib.url import canonicalize_url

    from ._params import _ParamParser

    class ScrapyZyteAPIRequestFingerprinter:
        @classmethod
        def from_crawler(cls, crawler):
            return cls(crawler)

        def __init__(self, crawler):
            settings = crawler.settings
            self._fallback_request_fingerprinter = create_instance(
                load_object(
                    settings.get(
                        "ZYTE_API_FALLBACK_REQUEST_FINGERPRINTER_CLASS",
                        REQUEST_FINGERPRINTER_CLASS,
                    )
                ),
                settings=crawler.settings,
                crawler=crawler,
            )
            self._cache: "WeakKeyDictionary[Request, bytes]" = WeakKeyDictionary()
            self._param_parser = _ParamParser(crawler, cookies_enabled=False)
            self._skip_keys = (
                "customHttpRequestHeaders",
                "echoData",
                "jobId",
                "requestHeaders",
                "experimental",
            )

        def _keep_fragments(self, api_params):
            return any(
                api_params.get(key, False) for key in ("browserHtml", "screenshot")
            )

        def fingerprint(self, request):
            if request in self._cache:
                return self._cache[request]
            api_params = self._param_parser.parse(request)
            if api_params is not None:
                api_params["url"] = canonicalize_url(
                    api_params["url"],
                    keep_fragments=self._keep_fragments(api_params),
                )
                for key in self._skip_keys:
                    api_params.pop(key, None)
                fingerprint_json = json.dumps(api_params, sort_keys=True)
                self._cache[request] = hashlib.sha1(fingerprint_json.encode()).digest()
                return self._cache[request]
            return self._fallback_request_fingerprinter.fingerprint(request)



from typing import Any, Callable, Dict, List, Sequence, Set, Type
from weakref import WeakKeyDictionary

from scrapy import Request
from scrapy.crawler import Crawler
from scrapy.utils.defer import maybe_deferred_to_future
from scrapy_poet import PageObjectInputProvider
from web_poet import BrowserHtml, BrowserResponse
from zyte_common_items import (
    Article,
    ArticleList,
    ArticleNavigation,
    Product,
    ProductList,
    ProductNavigation,
)

from scrapy_zyte_api.responses import ZyteAPITextResponse

try:
    # requires Scrapy >= 2.8
    from scrapy.http.request import NO_CALLBACK
except ImportError:
    NO_CALLBACK = None


class ZyteApiProvider(PageObjectInputProvider):
    name = "zyte_api"

    provided_classes = {
        BrowserResponse,
        BrowserHtml,
        Product,
        ProductList,
        ProductNavigation,
        Article,
        ArticleList,
        ArticleNavigation,
    }

    def __init__(self, injector):
        super().__init__(injector)
        self._cached_instances: WeakKeyDictionary[Request, Dict] = WeakKeyDictionary()

    def update_cache(self, request: Request, mapping: Dict[Type, Any]) -> None:
        if request not in self._cached_instances:
            self._cached_instances[request] = {}
        self._cached_instances[request].update(mapping)

    async def __call__(
        self, to_provide: Set[Callable], request: Request, crawler: Crawler
    ) -> Sequence[Any]:
        """Makes a Zyte API request to provide BrowserResponse and/or item dependencies."""
        # TODO what if ``response`` is already from Zyte API and contains something we need
        results: List[Any] = []

        for cls in list(to_provide):
            item = self._cached_instances.get(request, {}).get(cls)
            if item:
                results.append(item)
                to_provide.remove(cls)
        if not to_provide:
            return results

        html_requested = BrowserResponse in to_provide or BrowserHtml in to_provide
        item_keywords = {
            Product: "product",
            ProductList: "productList",
            ProductNavigation: "productNavigation",
            Article: "article",
            ArticleList: "articleList",
            ArticleNavigation: "articleNavigation",
        }

        zyte_api_meta = {}
        if html_requested:
            zyte_api_meta["browserHtml"] = True
        for item_type, kw in item_keywords.items():
            if item_type in to_provide:
                zyte_api_meta[kw] = True
        api_request = Request(
            url=request.url,
            meta={
                "zyte_api": zyte_api_meta,
                "zyte_api_default_params": False,
            },
            callback=NO_CALLBACK,
        )
        api_response: ZyteAPITextResponse = await maybe_deferred_to_future(
            crawler.engine.download(api_request)
        )

        assert api_response.raw_api_response
        if html_requested:
            html = BrowserHtml(api_response.raw_api_response["browserHtml"])
        else:
            html = None
        if BrowserHtml in to_provide:
            results.append(html)
            self.update_cache(request, {BrowserHtml: html})
        if BrowserResponse in to_provide:
            response = BrowserResponse(
                url=api_response.url,
                status=api_response.status,
                html=html,
            )
            results.append(response)
            self.update_cache(request, {BrowserResponse: response})
        for item_type, kw in item_keywords.items():
            if item_type in to_provide:
                item = item_type.from_dict(api_response.raw_api_response[kw])
                results.append(item)
                self.update_cache(request, {item_type: item})
        return results



from base64 import b64decode, b64encode
from copy import copy
from logging import getLogger
from os import environ
from typing import Any, Dict, List, Mapping, Optional, Set
from warnings import warn

from scrapy import Request
from scrapy.http.cookies import CookieJar
from scrapy.settings.default_settings import DEFAULT_REQUEST_HEADERS
from scrapy.settings.default_settings import USER_AGENT as DEFAULT_USER_AGENT

from ._cookies import _get_all_cookies

logger = getLogger(__name__)

_EXTRACT_KEYS = {
    "article",
    "articleList",
    "articleNavigation",
    "product",
    "productList",
    "productNavigation",
}
_BROWSER_KEYS = _EXTRACT_KEYS | {"browserHtml", "screenshot"}
_DEFAULT_API_PARAMS = {key: False for key in _BROWSER_KEYS}


def _iter_headers(
    *,
    api_params: Dict[str, Any],
    request: Request,
    header_parameter: str,
):
    headers = api_params.get(header_parameter)
    if headers not in (None, True):
        logger.warning(
            f"Request {request} defines the Zyte API {header_parameter} "
            f"parameter, overriding Request.headers. Use Request.headers "
            f"instead."
        )
        return
    if not request.headers:
        return
    for k, v in request.headers.items():
        if not v:
            continue
        decoded_v = b",".join(v).decode()
        lowercase_k = k.strip().lower()
        yield k, lowercase_k, decoded_v


def _map_custom_http_request_headers(
    *,
    api_params: Dict[str, Any],
    request: Request,
    skip_headers: Set[str],
):
    headers = []
    for k, lowercase_k, decoded_v in _iter_headers(
        api_params=api_params,
        request=request,
        header_parameter="customHttpRequestHeaders",
    ):
        if lowercase_k in skip_headers:
            if not (
                lowercase_k == b"cookie"
                or (lowercase_k == b"user-agent" and decoded_v == DEFAULT_USER_AGENT)
            ):
                logger.warning(
                    f"Request {request} defines header {k}, which "
                    f"cannot be mapped into the Zyte API "
                    f"customHttpRequestHeaders parameter."
                )
            continue
        headers.append({"name": k.decode(), "value": decoded_v})
    if headers:
        api_params["customHttpRequestHeaders"] = headers


def _map_request_headers(
    *,
    api_params: Dict[str, Any],
    request: Request,
    browser_headers: Dict[str, str],
):
    request_headers = {}
    for k, lowercase_k, decoded_v in _iter_headers(
        api_params=api_params,
        request=request,
        header_parameter="requestHeaders",
    ):
        key = browser_headers.get(lowercase_k)
        if key is not None:
            request_headers[key] = decoded_v
        elif not (
            (
                lowercase_k == b"accept"
                and decoded_v == DEFAULT_REQUEST_HEADERS["Accept"]
            )
            or (
                lowercase_k == b"accept-language"
                and decoded_v == DEFAULT_REQUEST_HEADERS["Accept-Language"]
            )
            or lowercase_k == b"cookie"
            or (lowercase_k == b"user-agent" and decoded_v == DEFAULT_USER_AGENT)
        ):
            logger.warning(
                f"Request {request} defines header {k}, which "
                f"cannot be mapped into the Zyte API requestHeaders "
                f"parameter."
            )
    if request_headers:
        api_params["requestHeaders"] = request_headers


def _set_request_headers_from_request(
    *,
    api_params: Dict[str, Any],
    request: Request,
    skip_headers: Set[str],
    browser_headers: Dict[str, str],
):
    """Updates *api_params*, in place, based on *request*."""
    custom_http_request_headers = api_params.get("customHttpRequestHeaders")
    request_headers = api_params.get("requestHeaders")
    response_body = api_params.get("httpResponseBody")

    if (
        response_body
        and custom_http_request_headers is not False
        or custom_http_request_headers is True
    ):
        _map_custom_http_request_headers(
            api_params=api_params,
            request=request,
            skip_headers=skip_headers,
        )
    elif custom_http_request_headers is False:
        api_params.pop("customHttpRequestHeaders")

    if (
        (not response_body or any(api_params.get(k) for k in _BROWSER_KEYS))
        and request_headers is not False
        or request_headers is True
    ):
        _map_request_headers(
            api_params=api_params,
            request=request,
            browser_headers=browser_headers,
        )
    elif request_headers is False:
        api_params.pop("requestHeaders")


def _set_http_response_body_from_request(
    *,
    api_params: Dict[str, Any],
    request: Request,
):
    if not any(api_params.get(k) for k in _BROWSER_KEYS):
        api_params.setdefault("httpResponseBody", True)
    elif api_params.get("httpResponseBody") is False:
        logger.warning(
            f"Request {request} unnecessarily defines the Zyte API "
            f"'httpResponseBody' parameter with its default value, False. "
            f"It will not be sent to the server."
        )
    if api_params.get("httpResponseBody") is False:
        api_params.pop("httpResponseBody")


def _set_http_response_headers_from_request(
    *,
    api_params: Dict[str, Any],
    default_params: Dict[str, Any],
    meta_params: Dict[str, Any],
):
    if api_params.get("httpResponseBody"):
        api_params.setdefault("httpResponseHeaders", True)
    elif (
        api_params.get("httpResponseHeaders") is False
        and not default_params.get("httpResponseHeaders") is False
    ):
        logger.warning(
            "You do not need to set httpResponseHeaders to False if "
            "neither httpResponseBody nor browserHtml are set to True. Note "
            "that httpResponseBody is set to True automatically if "
            "neither browserHtml nor screenshot are set to True."
        )
    if api_params.get("httpResponseHeaders") is False:
        api_params.pop("httpResponseHeaders")


def _set_http_response_cookies_from_request(
    *,
    api_params: Dict[str, Any],
):
    api_params.setdefault("experimental", {})
    api_params["experimental"].setdefault("responseCookies", True)
    if api_params["experimental"]["responseCookies"] is False:
        del api_params["experimental"]["responseCookies"]


def _set_http_request_cookies_from_request(
    *,
    api_params: Dict[str, Any],
    request: Request,
    cookie_jars: Dict[Any, CookieJar],
    max_cookies: int,
):
    api_params.setdefault("experimental", {})
    if "requestCookies" in api_params["experimental"]:
        request_cookies = api_params["experimental"]["requestCookies"]
        if request_cookies is False:
            del api_params["experimental"]["requestCookies"]
        elif not request_cookies and isinstance(request_cookies, list):
            logger.warning(
                (
                    "Request %(request)r is overriding automatic request "
                    "cookie mapping by explicitly setting "
                    "experimental.requestCookies to []. If this was your "
                    "intention, please use False instead of []. Otherwise, "
                    "stop defining experimental.requestCookies in your "
                    "request to let automatic mapping work."
                ),
                {
                    "request": request,
                },
            )
        return
    output_cookies = []
    input_cookies = _get_all_cookies(request, cookie_jars)
    input_cookie_count = len(input_cookies)
    if input_cookie_count > max_cookies:
        logger.warning(
            (
                "Request %(request)r would get %(count)r cookies, but request "
                "cookie automatic mapping is limited to %(max)r cookies "
                "(see the ZYTE_API_MAX_COOKIES setting), so only %(max)r "
                "cookies have been added to this request. To silence this "
                "warning, set the request cookies manually through the "
                "experimental.requestCookies Zyte API parameter instead. "
                "Alternatively, if Zyte API starts supporting more than "
                "%(max)r request cookies, update the ZYTE_API_MAX_COOKIES "
                "setting accordingly."
            ),
            {
                "request": request,
                "count": input_cookie_count,
                "max": max_cookies,
            },
        )
        input_cookies = input_cookies[:max_cookies]
    for input_cookie in input_cookies:
        output_cookie = {
            "name": input_cookie.name,
            "value": input_cookie.value,
            "domain": input_cookie.domain,
        }
        if input_cookie.path_specified:
            output_cookie["path"] = input_cookie.path
        output_cookies.append(output_cookie)
    if output_cookies:
        api_params["experimental"]["requestCookies"] = output_cookies


def _set_http_request_method_from_request(
    *,
    api_params: Dict[str, Any],
    request: Request,
):
    method = api_params.get("httpRequestMethod")
    if method:
        logger.warning(
            f"Request {request} uses the Zyte API httpRequestMethod "
            f"parameter, overriding Request.method. Use Request.method "
            f"instead."
        )
        if method != request.method:
            logger.warning(
                f"The HTTP method of request {request} ({request.method}) "
                f"does not match the Zyte API httpRequestMethod parameter "
                f"({method})."
            )
    elif request.method != "GET":
        api_params["httpRequestMethod"] = request.method


def _set_http_request_body_from_request(
    *,
    api_params: Dict[str, Any],
    request: Request,
):
    body = api_params.get("httpRequestBody")
    if body:
        logger.warning(
            f"Request {request} uses the Zyte API httpRequestBody parameter, "
            f"overriding Request.body. Use Request.body instead."
        )
        decoded_body = b64decode(body)
        if decoded_body != request.body:
            logger.warning(
                f"The body of request {request} ({request.body!r}) "
                f"does not match the Zyte API httpRequestBody parameter "
                f"({body!r}; decoded: {decoded_body!r})."
            )
    elif request.body != b"":
        base64_body = b64encode(request.body).decode()
        api_params["httpRequestBody"] = base64_body


def _unset_unneeded_api_params(
    *,
    api_params: Dict[str, Any],
    default_params: Dict[str, Any],
    request: Request,
):
    for param, default_value in _DEFAULT_API_PARAMS.items():
        if api_params.get(param) != default_value:
            continue
        if param not in default_params or default_params.get(param) == default_value:
            logger.warning(
                f"Request {request} unnecessarily defines the Zyte API {param!r} "
                f"parameter with its default value, {default_value!r}. It will "
                f"not be sent to the server."
            )
        api_params.pop(param)


def _update_api_params_from_request(
    api_params: Dict[str, Any],
    request: Request,
    *,
    default_params: Dict[str, Any],
    meta_params: Dict[str, Any],
    skip_headers: Set[str],
    browser_headers: Dict[str, str],
    cookies_enabled: bool,
    cookie_jars: Optional[Dict[Any, CookieJar]],
    max_cookies: int,
):
    _set_http_response_body_from_request(api_params=api_params, request=request)
    _set_http_response_headers_from_request(
        api_params=api_params,
        default_params=default_params,
        meta_params=meta_params,
    )
    _set_http_request_method_from_request(api_params=api_params, request=request)
    _set_request_headers_from_request(
        api_params=api_params,
        request=request,
        skip_headers=skip_headers,
        browser_headers=browser_headers,
    )
    _set_http_request_body_from_request(api_params=api_params, request=request)
    if cookies_enabled:
        assert cookie_jars is not None  # typing
        _set_http_response_cookies_from_request(api_params=api_params)
        _set_http_request_cookies_from_request(
            api_params=api_params,
            request=request,
            cookie_jars=cookie_jars,
            max_cookies=max_cookies,
        )
        if not api_params["experimental"]:
            del api_params["experimental"]
    _unset_unneeded_api_params(
        api_params=api_params, request=request, default_params=default_params
    )
    return api_params


def _copy_meta_params_as_dict(
    meta_params: Dict[str, Any],
    *,
    param: str,
    request: Request,
):
    if meta_params is True:
        return {}
    elif not isinstance(meta_params, Mapping):
        raise ValueError(
            f"'{param}' parameters in the request meta should be provided as "
            f"a dictionary, got {type(meta_params)} instead in {request}."
        )
    else:
        return copy(meta_params)


def _merge_params(
    *,
    default_params: Dict[str, Any],
    meta_params: Dict[str, Any],
    param: str,
    setting: str,
    request: Request,
    context: Optional[List[str]] = None,
):
    params = copy(default_params)
    meta_params = copy(meta_params)
    context = context or []
    for k in list(meta_params):
        if isinstance(meta_params[k], dict):
            meta_params[k] = _merge_params(
                default_params=params.get(k, {}),
                meta_params=meta_params[k],
                param=param,
                setting=setting,
                request=request,
                context=context + [k],
            )
        if meta_params[k] not in (None, {}):
            continue
        meta_params.pop(k)
        if k in params:
            params.pop(k)
        else:
            qual_param = ".".join(context + [k])
            logger.warning(
                f"In request {request} {param!r} parameter {qual_param} is "
                f"None, which is a value reserved to unset parameters defined "
                f"in the {setting} setting, but the setting does not define "
                f"such a parameter."
            )
    params.update(meta_params)
    return params


def _get_raw_params(
    request: Request,
    *,
    default_params: Dict[str, Any],
):
    meta_params = request.meta.get("zyte_api", False)
    if meta_params is False:
        return None

    if not meta_params and meta_params != {}:
        warn(
            f"Setting the zyte_api request metadata key to "
            f"{meta_params!r} is deprecated. Use False instead.",
            DeprecationWarning,
        )
        return None

    meta_params = _copy_meta_params_as_dict(
        meta_params,
        param="zyte_api",
        request=request,
    )

    return _merge_params(
        default_params=default_params,
        meta_params=meta_params,
        param="zyte_api",
        setting="ZYTE_API_DEFAULT_PARAMS",
        request=request,
    )


def _get_automap_params(
    request: Request,
    *,
    default_enabled: bool,
    default_params: Dict[str, Any],
    skip_headers: Set[str],
    browser_headers: Dict[str, str],
    cookies_enabled: bool,
    cookie_jars: Optional[Dict[Any, CookieJar]],
    max_cookies: int,
):
    meta_params = request.meta.get("zyte_api_automap", default_enabled)
    if meta_params is False:
        return None

    meta_params = _copy_meta_params_as_dict(
        meta_params,
        param="zyte_api_automap",
        request=request,
    )

    params = _merge_params(
        default_params=default_params,
        meta_params=meta_params,
        param="zyte_api_automap",
        setting="ZYTE_API_AUTOMAP_PARAMS",
        request=request,
    )

    _update_api_params_from_request(
        params,
        request,
        default_params=default_params,
        meta_params=meta_params,
        skip_headers=skip_headers,
        browser_headers=browser_headers,
        cookies_enabled=cookies_enabled,
        cookie_jars=cookie_jars,
        max_cookies=max_cookies,
    )

    return params


def _get_api_params(
    request: Request,
    *,
    default_params: Dict[str, Any],
    transparent_mode: bool,
    automap_params: Dict[str, Any],
    skip_headers: Set[str],
    browser_headers: Dict[str, str],
    job_id: Optional[str],
    cookies_enabled: bool,
    cookie_jars: Optional[Dict[Any, CookieJar]],
    max_cookies: int,
) -> Optional[dict]:
    """Returns a dictionary of API parameters that must be sent to Zyte API for
    the specified request, or None if the request should not be sent through
    Zyte API."""
    api_params = _get_raw_params(request, default_params=default_params)
    if api_params is None:
        api_params = _get_automap_params(
            request,
            default_enabled=transparent_mode,
            default_params=automap_params,
            skip_headers=skip_headers,
            browser_headers=browser_headers,
            cookies_enabled=cookies_enabled,
            cookie_jars=cookie_jars,
            max_cookies=max_cookies,
        )
        if api_params is None:
            return None
    elif request.meta.get("zyte_api_automap", False) is not False:
        raise ValueError(
            f"Request {request} combines manually-defined parameters and "
            f"automatically-mapped parameters."
        )

    if job_id is not None:
        api_params["jobId"] = job_id

    api_params["url"] = request.url

    return api_params


def _load_default_params(settings, setting):
    params = settings.getdict(setting)
    for param in list(params):
        if params[param] not in (None, {}):
            continue
        logger.warning(
            f"Parameter {param!r} in the {setting} setting is "
            f"{params[param]!r}. Default parameters should never be "
            f"{params[param]!r}."
        )
        params.pop(param)
    return params


def _load_skip_headers(settings):
    return {
        header.strip().lower().encode()
        for header in settings.getlist(
            "ZYTE_API_SKIP_HEADERS",
            ["Cookie", "User-Agent"],
        )
    }


def _load_browser_headers(settings):
    browser_headers = settings.getdict(
        "ZYTE_API_BROWSER_HEADERS",
        {"Referer": "referer"},
    )
    return {k.strip().lower().encode(): v for k, v in browser_headers.items()}


class _ParamParser:
    def __init__(self, crawler, cookies_enabled=None):
        settings = crawler.settings
        self._automap_params = _load_default_params(settings, "ZYTE_API_AUTOMAP_PARAMS")
        self._browser_headers = _load_browser_headers(settings)
        self._default_params = _load_default_params(settings, "ZYTE_API_DEFAULT_PARAMS")
        self._job_id = environ.get("SHUB_JOBKEY", None)
        self._transparent_mode = settings.getbool("ZYTE_API_TRANSPARENT_MODE", False)
        self._skip_headers = _load_skip_headers(settings)
        self._warn_on_cookies = False
        if cookies_enabled is not None:
            self._cookies_enabled = cookies_enabled
        elif settings.getbool("ZYTE_API_EXPERIMENTAL_COOKIES_ENABLED") is True:
            self._cookies_enabled = settings.getbool("COOKIES_ENABLED")
            if not self._cookies_enabled:
                logger.warning(
                    "ZYTE_API_EXPERIMENTAL_COOKIES_ENABLED is True, but it "
                    "will have no effect because COOKIES_ENABLED is False."
                )
        else:
            self._cookies_enabled = False
            self._warn_on_cookies = settings.getbool("COOKIES_ENABLED")
        self._max_cookies = settings.getint("ZYTE_API_MAX_COOKIES", 100)
        self._crawler = crawler
        self._cookie_jars = None

    def parse(self, request):
        dont_merge_cookies = request.meta.get("dont_merge_cookies", False)
        use_default_params = request.meta.get("zyte_api_default_params", True)
        cookies_enabled = self._cookies_enabled and not dont_merge_cookies
        params = _get_api_params(
            request,
            default_params=self._default_params if use_default_params else {},
            transparent_mode=self._transparent_mode,
            automap_params=self._automap_params,
            skip_headers=self._skip_headers,
            browser_headers=self._browser_headers,
            job_id=self._job_id,
            cookies_enabled=cookies_enabled,
            cookie_jars=self._cookie_jars,
            max_cookies=self._max_cookies,
        )
        if not dont_merge_cookies and self._warn_on_cookies:
            self._handle_warn_on_cookies(request, params)
        return params

    def _handle_warn_on_cookies(self, request, params):
        if params and params.get("experimental", {}).get("requestCookies") is not None:
            return
        if self._cookie_jars is None:
            return
        input_cookies = _get_all_cookies(request, self._cookie_jars)
        if len(input_cookies) <= 0:
            return
        logger.warning(
            (
                "Cookies are enabled for request %(request)r, and there are "
                "cookies in the cookiejar, but "
                "ZYTE_API_EXPERIMENTAL_COOKIES_ENABLED is False, so automatic "
                "mapping will not map cookies for this or any other request. "
                "To silence this warning, disable cookies for all requests "
                "that use automated mapping, either with the "
                "COOKIES_ENABLED setting or with the dont_merge_cookies "
                "request metadata key."
            ),
            {
                "request": request,
            },
        )
        self._warn_on_cookies = False



import scrapy
from packaging.version import Version

_SCRAPY_VERSION = Version(scrapy.__version__)
_SCRAPY_2_1_0 = Version("2.1.0")
_SCRAPY_2_4_0 = Version("2.4.0")
_SCRAPY_2_5_0 = Version("2.5.0")
_SCRAPY_2_6_0 = Version("2.6.0")

# Need to install an asyncio reactor before download handler imports to work
# around:
# https://github.com/scrapy/scrapy/commit/0946eb335a285e1f210ba1185a564699f53b17d8
# Fixed in:
# https://github.com/scrapy/scrapy/commit/e4bdd1cb958b7d89b86ea66f0af1cec2d91a6d44
_NEEDS_EARLY_REACTOR = _SCRAPY_2_4_0 <= _SCRAPY_VERSION < _SCRAPY_2_6_0

_RESPONSE_HAS_ATTRIBUTES = _SCRAPY_VERSION >= _SCRAPY_2_6_0
_RESPONSE_HAS_IP_ADDRESS = _SCRAPY_VERSION >= _SCRAPY_2_1_0
_RESPONSE_HAS_PROTOCOL = _SCRAPY_VERSION >= _SCRAPY_2_5_0



import setuptools

setuptools.setup(
    name="scrapy-zyte-api",
    version="0.11.0",
    description="Client library to process URLs through Zyte API",
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    author="Zyte Group Ltd",
    author_email="info@zyte.com",
    url="https://github.com/scrapy-plugins/scrapy-zyte-api",
    packages=["scrapy_zyte_api"],
    install_requires=[
        "packaging>=20.0",
        "scrapy>=2.0.1",
        "zyte-api>=0.4.0",
    ],
    extras_require={
        "provider": [
            "scrapy-poet>=0.11.0",
            "web-poet>=0.13.0",
            "zyte-common-items>=0.7.0",
        ]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)
